import struct
from crccheck.crc import Crc32Mpeg2 as CRC32
import serial
import time
from packaging.version import parse as parse_version
import requests
import hashlib
import tempfile
from stm32loader.main import main as stm32loader_main
import enum
from smd.SMD_device import *


# enter here for extra commands: 
#class Device_ExtraCommands(enum.IntEnum):
	# .......... start with 11
	# .......... end of extra commmands max: 39


Index_Green = enum.IntEnum('Index', [
	'Header',
	'DeviceID',
	'DeviceFamily',
	'PackageSize',
	'Command',
	'Status',
	'HardwareVersion',
	'SoftwareVersion',
	'Baudrate', #'WritableStart' = iBaudrate
	# user parameter start
	'OperationMode',
	'Enable',
	'Vbus_read',
	'Temprature_read',
	'currentId_loop_kp',
	'currentId_loop_ki',
	'currentId_loop_kd',
	'currentIq_loop_kp',
	'currentIq_loop_ki',
	'currentIq_loop_kd',
	'velocity_loop_kp',
	'velocity_loop_ki',
	'velocity_loop_kd',
	'position_loop_kp',
	'position_loop_ki',
	'position_loop_kd',
	'max_position',
	'min_position',
	'max_velocity',
	'max_current',
	'current_Va',
	'current_Vb',
	'current_Vc',
	'current_Ia',
	'current_Ib',
	'current_Ic',
	'current_Id',
	'current_Iq',
	'current_velocity',
	'current_position',
	'current_electrical_degree',
	'current_electrical_radian',
	'setpoint_current',
	'setpoint_velocity',
	'setpoint_position',
	'openloop_voltage_size',
	'openloop_angle_degree',
	'current_lock_angle_degree',
	# user parameter end
	'Config_TimeStamp',
	'Config_Description',
	'CRCValue',
], start=0)



def scan_blue_devices(port:SerialPort):
	device = Blue(0, port)
	available_devices = []

	for id in range(0,255):
		device._id = id
		if(device.ping()):
			available_devices.append(id)

	return available_devices


class Green(SMD_Device):
	_PRODUCT_TYPE = 0xDA
	_PACKAGE_ESSENTIAL_SIZE = 6
	_STATUS_KEY_LIST = ['EEPROM', 'Software Version', 'Hardware Version']
	__RELEASE_URL = "https://api.github.com/repos/AAcrome-Smart-Motion-Devices/SMD-Blue-Firmware/releases/{version}"

	class Operation_Mode():
		OPERATION_OPENLOOP = 0
		OPERATION_CURRENT_LOCK = 1
		OPERATION_FOC_TORQUE = 2
		OPERATION_FOC_VELOCITY = 3
		OPERATION_FOC_POSITION = 4

	def __init__(self, ID, port:SerialPort) -> bool:
		self.__ack_size = 0
		if ID > 254 or ID < 0:
			raise ValueError("Device ID can not be higher than 254 or lower than 0!")
		device_special_data = [
            Data_(Index_Green.Header, 'B', False, 0x55),
            Data_(Index_Green.DeviceID, 'B'),
			Data_(Index_Green.DeviceFamily, 'B'),
            Data_(Index_Green.PackageSize, 'B'),
            Data_(Index_Green.Command, 'B'),
			Data_(Index_Green.Status, 'B'),
            Data_(Index_Green.HardwareVersion, 'I'),
            Data_(Index_Green.SoftwareVersion, 'I'),
            Data_(Index_Green.Baudrate, 'I'),
			# user parameter starts
			Data_(Index_Green.OperationMode, 'B'),
			Data_(Index_Green.Enable, 'B'),
			Data_(Index_Green.Vbus_read, 'f'),
			Data_(Index_Green.Temprature_read, 'f'),
			Data_(Index_Green.currentId_loop_kp, 'f'),
			Data_(Index_Green.currentId_loop_ki, 'f'),
			Data_(Index_Green.currentId_loop_kd, 'f'),
			Data_(Index_Green.currentIq_loop_kp, 'f'),
			Data_(Index_Green.currentIq_loop_ki, 'f'),
			Data_(Index_Green.currentIq_loop_kd, 'f'),
			Data_(Index_Green.velocity_loop_kp, 'f'),
			Data_(Index_Green.velocity_loop_ki, 'f'),
			Data_(Index_Green.velocity_loop_kd, 'f'),
			Data_(Index_Green.position_loop_kp, 'f'),
			Data_(Index_Green.position_loop_ki, 'f'),
			Data_(Index_Green.position_loop_kd, 'f'),
			Data_(Index_Green.max_position, 'i'),
			Data_(Index_Green.min_position, 'i'),
			Data_(Index_Green.max_velocity, 'f'),
			Data_(Index_Green.max_current, 'f'),
			Data_(Index_Green.current_Va, 'f'),
			Data_(Index_Green.current_Vb, 'f'),
			Data_(Index_Green.current_Vc, 'f'),
			Data_(Index_Green.current_Ia, 'f'),
			Data_(Index_Green.current_Ib, 'f'),
			Data_(Index_Green.current_Ic, 'f'),
			Data_(Index_Green.current_Id, 'f'),
			Data_(Index_Green.current_Iq, 'f'),
			Data_(Index_Green.current_velocity, 'f'),
			Data_(Index_Green.current_position, 'i'),
			Data_(Index_Green.current_electrical_degree, 'f'),
			Data_(Index_Green.current_electrical_radian, 'f'),
			Data_(Index_Green.setpoint_current, 'f'),
			Data_(Index_Green.setpoint_velocity, 'f'),
			Data_(Index_Green.setpoint_position, 'f'),
			Data_(Index_Green.openloop_voltage_size, 'f'),
			Data_(Index_Green.openloop_angle_degree, 'f'),
			Data_(Index_Green.current_lock_angle_degree, 'f'),
			# user parameter end
			Data_(Index_Green.Config_TimeStamp, 'Q'),
			Data_(Index_Green.Config_Description, '100s'),
            Data_(Index_Green.CRCValue, 'I'),
        ]
		super().__init__(ID, self._PRODUCT_TYPE, device_special_data, port)
		self._vars[Index_Green.DeviceID].value(ID)

	# user start for extra commands.
	#def command(self): 

	def __del__(self):
		pass

	def get_latest_fw_version(self):
		""" Get the latest firmware version from the Github servers.

		Returns:
			String: Latest firmware version
		"""
		response = requests.get(url=self.__class__.__RELEASE_URL.format(version='latest'))
		if (response.status_code in [200, 302]):
			return (response.json()['tag_name'])

	def update_fw_version(self, version=''):
		""" Update firmware version with respect to given version string.

		Args:
			id (int): The device ID of the driver
			version (str, optional): Desired firmware version. Defaults to ''.

		Returns:
			Bool: True if the firmware is updated
		"""

		fw_file = tempfile.NamedTemporaryFile("wb+",delete=False)
		if version == '':
			version = 'latest'
		else:
			version = 'tags/' + version

		response = requests.get(url=self.__class__.__RELEASE_URL.format(version=version))
		if response.status_code in [200, 302]:
			assets = response.json()['assets']

			fw_dl_url = None
			md5_dl_url = None
			for asset in assets:
				if '.bin' in asset['name']:
					fw_dl_url = asset['browser_download_url']
				elif '.md5' in asset['name']:
					md5_dl_url = asset['browser_download_url']

			if None in [fw_dl_url, md5_dl_url]:
				raise Exception("Could not found requested firmware file! Check your connection to GitHub.")

			#  Get binary firmware file
			md5_fw = None
			response = requests.get(fw_dl_url, stream=True)
			if (response.status_code in [200, 302]):
				fw_file.write(response.content)
				md5_fw = hashlib.md5(response.content).hexdigest()
			else:
				raise Exception("Could not fetch requested binary file! Check your connection to GitHub.")

			#  Get MD5 file
			response = requests.get(md5_dl_url, stream=True)
			if (response.status_code in [200, 302]):
				md5_retreived = response.text.split(' ')[0]
				if (md5_fw == md5_retreived):

					# Put the driver in to bootloader mode
					self.enter_bootloader()
					time.sleep(0.1)

					# Close serial port
					serial_settings = self._port._ph.get_settings()
					self._port._ph.close()

					# Upload binary
					args = ['-p', self._port._ph.portstr, '-b', str(115200), '-e', '-w', '-v', fw_file.name]
					stm32loader_main(*args)

					# Delete uploaded binary
					if (not fw_file.closed):
						fw_file.close()

					# Re open port to the user with saved settings
					self._port._ph.apply_settings(serial_settings)
					self._port._ph.open()
					return True

				else:
					raise Exception("MD5 Mismatch!")
			else:
				raise Exception("Could not fetch requested MD5 file! Check your connection to GitHub.")
		else:
			raise Exception("Could not found requested firmware files list! Check your connection to GitHub.")
		
	def enable_torque(self, en: bool):
		""" Enable power to the motor of the driver.

    	Args:
    	    id (int): The device ID of the driver
    	    en (bool): Enable. True enables the torque.
    	"""

		self.set_variables([Index_Green.Enable, en])
		self._post_sleep()

	def set_config_timeStamp(self):
		epoch_seconds = int(time.time())
		self.set_variables([Index_Green.Config_TimeStamp, epoch_seconds])
		self._post_sleep()
		
	def set_config_description(self, description:str):
		if len(description) >= 100:
			text = description[:99] + '\0'
		else:
			text = description + '\0'
			text = text.ljust(100, ' ')
		text = text.encode('ascii')  # veya utf-8 eğer uyumluysa

		self.set_variables([Index_Green.Config_Description, text])
		self._post_sleep()



	def get_FOC_parameters(self, package_number:int):
		if package_number >= 4:
			raise "invalid package number ex: 0, 1, 2"
		classic_package = [
			Index_Green.Enable,
			Index_Green.current_Id, Index_Green.current_Iq,
			Index_Green.current_velocity, Index_Green.current_position,
			Index_Green.Temprature_read,
			Index_Green.setpoint_current, Index_Green.setpoint_velocity, Index_Green.setpoint_position
		]
		package_0 = [
			Index_Green.currentId_loop_kp, 
			Index_Green.currentId_loop_ki, 
			Index_Green.currentId_loop_kd, 
			Index_Green.currentIq_loop_kp, 
			Index_Green.currentIq_loop_ki, 
			Index_Green.currentIq_loop_kd, 
		]
		package_1 = [
			Index_Green.velocity_loop_kp,
			Index_Green.velocity_loop_ki,
			Index_Green.velocity_loop_kd,
		]
		package_2 = [
			Index_Green.position_loop_kp,
			Index_Green.position_loop_ki,
			Index_Green.position_loop_kd,
		]

		if package_number == 0:
			return self.get_variables(*classic_package , *package_0)
			
		elif package_number == 1:
			return self.get_variables(*classic_package , *package_1)
			
		elif package_number == 2:
			return self.get_variables(*classic_package , *package_2)
		
		elif package_number == 3:
			return self.get_variables(*classic_package)
		


import struct
from crccheck.crc import Crc32Mpeg2 as CRC32
import serial
import time
from packaging.version import parse as parse_version
import requests
import hashlib
import tempfile
from stm32loader.main import main as stm32loader_main
import enum
from smd.SMD_device import *

class InvalidIndexError(BaseException):
    pass

class UnsupportedHardware(BaseException):
    pass

class UnsupportedFirmware(BaseException):
    pass

# enter here for extra commands: 
class ExtraCommands_Red(enum.IntEnum):
    RESET_ENC = 0x06
    TUNE = 0x07
    ERROR_CLEAR = 0x18
    SYNC_WRITE = 0x40 | 0x01


Index_Red = enum.IntEnum('Index', [
	'Header',
	'DeviceID',
	'DeviceFamily',
	'PackageSize',
	'Command',
	'Status',
	'HardwareVersion',
	'SoftwareVersion',
	'Baudrate',
	# user parameter start
    'OperationMode',
    'TorqueEnable',
    'OutputShaftCPR',
    'OutputShaftRPM',
    'UserIndicator',
    'MinimumPositionLimit',
    'MaximumPositionLimit',
    'TorqueLimit',
    'VelocityLimit',
    'PositionFF',
    'VelocityFF',
    'TorqueFF',
    'PositionDeadband',
    'VelocityDeadband',
    'TorqueDeadband',
    'PositionOutputLimit',
    'VelocityOutputLimit',
    'TorqueOutputLimit',
    'PositionScalerGain',
    'PositionPGain',
    'PositionIGain',
    'PositionDGain',
    'VelocityScalerGain',
    'VelocityPGain',
    'VelocityIGain',
    'VelocityDGain',
    'TorqueScalerGain',
    'TorquePGain',
    'TorqueIGain',
    'TorqueDGain',
    'SetPosition',
    'PositionControlMode',
	'SCurveSetpoint',
	'ScurveAccel',
	'SCurveMaxVelocity',
	'SCurveTime',
    'SetVelocity',
    'SetVelocityAcceleration',
    'SetTorque',
    'SetDutyCycle',
    'PresentPosition',              
    'PresentVelocity',
    'MotorCurrent',
    'AnalogPort',
	# user parameter end
	'CRCValue',
], start=0)

class OperationMode_Red():
    PWM = 0
    Position = 1
    Velocity = 2
    Torque = 3


class Red(SMD_Device):
    _PRODUCT_TYPE = 0xBA
    _PACKAGE_ESSENTIAL_SIZE = 6
    _STATUS_KEY_LIST = ['EEPROM', 'Software Version', 'Hardware Version']
    __RELEASE_URL = "https://api.github.com/repos/Acrome-Smart-Motion-Devices/SMD-Red-Firmware/releases/{version}"

    def __init__(self, ID, port:SerialPort) -> bool:
		
        self.__ack_size = 0
        self._config = None
        self._fw_file = None
        if ID > 254 or ID < 0:
            raise ValueError("Device ID can not be higher than 254 or lower than 0!")
        device_special_data = [
            Data_(Index_Red.Header, 'B', False, 0x55),
            Data_(Index_Red.DeviceID, 'B'),
            Data_(Index_Red.DeviceFamily, 'B', False, self.__class__._PRODUCT_TYPE),
            Data_(Index_Red.PackageSize, 'B'),
            Data_(Index_Red.Command, 'B'),
            Data_(Index_Red.Status, 'B'),
            Data_(Index_Red.HardwareVersion, 'I'),
            Data_(Index_Red.SoftwareVersion, 'I'),
            Data_(Index_Red.Baudrate, 'I'),
            # user parameter starts
            Data_(Index_Red.OperationMode, 'B'),
            Data_(Index_Red.TorqueEnable, 'B'),
            Data_(Index_Red.OutputShaftCPR, 'f'),
            Data_(Index_Red.OutputShaftRPM, 'f'),
            Data_(Index_Red.UserIndicator, 'B'),
            Data_(Index_Red.MinimumPositionLimit, 'i'),
            Data_(Index_Red.MaximumPositionLimit, 'i'),
            Data_(Index_Red.TorqueLimit, 'H'),
            Data_(Index_Red.VelocityLimit, 'H'),
            Data_(Index_Red.PositionFF, 'f'),
            Data_(Index_Red.VelocityFF, 'f'),
            Data_(Index_Red.TorqueFF, 'f'),
            Data_(Index_Red.PositionDeadband, 'f'),
            Data_(Index_Red.VelocityDeadband, 'f'),
            Data_(Index_Red.TorqueDeadband, 'f'),
            Data_(Index_Red.PositionOutputLimit, 'f'),
            Data_(Index_Red.VelocityOutputLimit, 'f'),
            Data_(Index_Red.TorqueOutputLimit, 'f'),
            Data_(Index_Red.PositionScalerGain, 'f'),
            Data_(Index_Red.PositionPGain, 'f'),
            Data_(Index_Red.PositionIGain, 'f'),
            Data_(Index_Red.PositionDGain, 'f'),
            Data_(Index_Red.VelocityScalerGain, 'f'),
            Data_(Index_Red.VelocityPGain, 'f'),
            Data_(Index_Red.VelocityIGain, 'f'),
            Data_(Index_Red.VelocityDGain, 'f'),
            Data_(Index_Red.TorqueScalerGain, 'f'),
            Data_(Index_Red.TorquePGain, 'f'),
            Data_(Index_Red.TorqueIGain, 'f'),
            Data_(Index_Red.TorqueDGain, 'f'),
            Data_(Index_Red.SetPosition, 'f'),
            Data_(Index_Red.PositionControlMode, 'B'),      # S Curve Position Control / 1 is SCurve(goTo function) 0 is direct control.
            Data_(Index_Red.SCurveSetpoint, 'f'),
            Data_(Index_Red.ScurveAccel, 'f'),
            Data_(Index_Red.SCurveMaxVelocity, 'f'),
            Data_(Index_Red.SCurveTime, 'f'),
            Data_(Index_Red.SetVelocity, 'f'),
            Data_(Index_Red.SetVelocityAcceleration, 'f'),
            Data_(Index_Red.SetTorque, 'f'),
            Data_(Index_Red.SetDutyCycle, 'f'),
            Data_(Index_Red.PresentPosition, 'f'),
            Data_(Index_Red.PresentVelocity, 'f'),
            Data_(Index_Red.MotorCurrent, 'f'),
            Data_(Index_Red.AnalogPort, 'H'),
            # user parameter end			
            Data_(Index_Red.CRCValue, 'I'),
        ]
        super().__init__(ID, self._PRODUCT_TYPE, device_special_data, port)
        self._vars[Index_Red.DeviceID].value(ID)

    def __del__(self):
        pass

    def get_latest_fw_version(self):
        """ Get the latest firmware version from the Github servers.

        Returns:
            String: Latest firmware version
        """
        response = requests.get(url=self.__class__.__RELEASE_URL.format(version='latest'))
        if (response.status_code in [200, 302]):
            return (response.json()['tag_name'])

    def update_fw_version(self, version=''):
        """ Update firmware version with respect to given version string.

        Args:
            id (int): The device ID of the driver
            version (str, optional): Desired firmware version. Defaults to ''.

        Returns:
            Bool: True if the firmware is updated
        """

        fw_file = tempfile.NamedTemporaryFile("wb+",delete=False)
        if version == '':
            version = 'latest'
        else:
            version = 'tags/' + version

        response = requests.get(url=self.__class__.__RELEASE_URL.format(version=version))
        if response.status_code in [200, 302]:
            assets = response.json()['assets']

            fw_dl_url = None
            md5_dl_url = None
            for asset in assets:
                if '.bin' in asset['name']:
                    fw_dl_url = asset['browser_download_url']
                elif '.md5' in asset['name']:
                    md5_dl_url = asset['browser_download_url']

            if None in [fw_dl_url, md5_dl_url]:
                raise Exception("Could not found requested firmware file! Check your connection to GitHub.")

            #  Get binary firmware file
            md5_fw = None
            response = requests.get(fw_dl_url, stream=True)
            if (response.status_code in [200, 302]):
                fw_file.write(response.content)
                md5_fw = hashlib.md5(response.content).hexdigest()
            else:
                raise Exception("Could not fetch requested binary file! Check your connection to GitHub.")

            #  Get MD5 file
            response = requests.get(md5_dl_url, stream=True)
            if (response.status_code in [200, 302]):
                md5_retreived = response.text.split(' ')[0]
                if (md5_fw == md5_retreived):

                    # Put the driver in to bootloader mode
                    self.enter_bootloader()
                    time.sleep(0.1)

                    # Close serial port
                    serial_settings = self.__ph.get_settings()
                    self.__ph.close()

                    # Upload binary
                    args = ['-p', self.__ph.portstr, '-b', str(115200), '-e', '-w', '-v', fw_file.name]
                    stm32loader_main(*args)

                    # Delete uploaded binary
                    if (not fw_file.closed):
                        fw_file.close()

                    # Re open port to the user with saved settings
                    self.__ph.apply_settings(serial_settings)
                    self.__ph.open()
                    return True

                else:
                    raise Exception("MD5 Mismatch!")
            else:
                raise Exception("Could not fetch requested MD5 file! Check your connection to GitHub.")
        else:
            raise Exception("Could not found requested firmware files list! Check your connection to GitHub.")

    def update_driver_baudrate(self, br: int):
        """Update the baudrate of the driver with
        given device ID. Following the method, the master
        baudrate must be updated accordingly to initiate a
        communication line with the board.

        Args:
            id (int): The device ID of the driver
            br (int): New baudrate value

        Raises:
            ValueError: Baudrate is not valid
        """

        if (br < 3053) or (br > 12500000):
            raise ValueError("{br} is not in acceptable range!")

        self.set_variables([Index_Red.Baudrate, br])
        self._post_sleep()
        self.eeprom_save()
        self._post_sleep()
        self.reboot(id)
        self._init_sleep()

    def get_driver_baudrate(self):
        """ Get the current baudrate from the driver.

        Args:
            id (int): The device ID of the driver.

        Returns:
            list | None: Returns the list containing the baudrate, otherwise None.
        """
        return self.get_variables(Index_Red.Baudrate)[0]
    
    def reset_encoder(self): #burayi kontrol et.
        """ Reset the encoder.

        Args:
            id (int): The device ID of the driver.
        """
        self.pureCommand(command_number= ExtraCommands_Red.RESET_ENC)
        self._post_sleep()

    def enable_torque(self, en: bool):
        """ Enable power to the motor of the driver.

        Args:
            id (int): The device ID of the driver
            en (bool): Enable. True enables the torque.
        """

        self.set_variables([Index_Red.TorqueEnable, en])
        self._post_sleep()

    def pid_tuner(self):
        """ Start PID auto-tuning routine. This routine will estimate
        PID coefficients for position and velocity control operation modes.

        Args:
            id (int): The device ID of the driver.
        """
        self.pureCommand(command_number= ExtraCommands_Red.TUNE)
        self._post_sleep()

    def set_operation_mode(self, mode:OperationMode_Red):
        """ Set the operation mode of the driver.

        Args:
            id (int): The device ID of the driver.
            mode (OperationMode): One of the PWM, Position, Velocity, Torque modes.
        """

        self.set_variables([Index_Red.OperationMode, mode])
        self._post_sleep()

    def get_operation_mode(self):
        """ Get the current operation mode from the driver.

        Args:
            id (int): The device ID of the driver.

        Returns:
            list | None: Returns the list containing the operation mode, otherwise None.
        """
        return self.get_variables(Index_Red.OperationMode)[0]

    def set_shaft_cpr(self, cpr: float):
        """ Set the count per revolution (CPR) of the motor output shaft.

        Args:
            id (int): The device ID of the driver.
            cpr (float): The CPR value of the output shaft/
        """
        self.set_variables([Index_Red.OutputShaftCPR, cpr])
        self._post_sleep()

    def get_shaft_cpr(self):
        """ Get the count per revolution (CPR) of the motor output shaft.

        Args:
            id (int): The device ID of the driver.

        Returns:
            list | None: Returns the list containing the output shaft CPR, otherwise None.
        """
        return self.get_variables(Index_Red.OutputShaftCPR)[0]

    def set_shaft_rpm(self, rpm: float):
        """ Set the revolution per minute (RPM) value of the output shaft at 12V rating.

        Args:
            id (int): The device ID of the driver.
            rpm (float): The RPM value of the output shaft at 12V
        """
        self.set_variables([Index_Red.OutputShaftRPM, rpm])
        self._post_sleep()

    def get_shaft_rpm(self):
        """ Get the revolution per minute (RPM) value of the output shaft at 12V rating.

        Args:
            id (int): The device ID of the driver.

        Returns:
            list | None: Returns the list containing the output shaft RPM characteristics, otherwise None.
        """
        return self.get_variables(Index_Red.OutputShaftRPM)

    def set_user_indicator(self):
        """ Set the user indicator color for 5 seconds. The user indicator color is cyan.

        Args:
            id (int): The device ID of the driver.
        """
        self.set_variables([Index_Red.UserIndicator, 1])
        self._post_sleep()

    def set_position_limits(self, plmin: int, plmax: int):
        """ Set the position limits of the motor in terms of encoder ticks.
        Default for min is -2,147,483,648 and for max is 2,147,483,647.
        The torque ise disabled if the value is exceeded so a tolerence
        factor should be taken into consideration when setting this values. 

        Args:
            id (int): The device ID of the driver.
            plmin (int): The minimum position limit.
            plmax (int): The maximum position limit.
        """
        self.set_variables([Index_Red.MinimumPositionLimit, plmin], [Index_Red.MaximumPositionLimit, plmax])
        self._post_sleep()

    def get_position_limits(self):
        """ Get the position limits of the motor in terms of encoder ticks.

        Args:
            id (int): The device ID of the driver.

        Returns:
            list | None: Returns the list containing the position limits, otherwise None.
        """
        return self.get_variables(Index_Red.MinimumPositionLimit, Index_Red.MaximumPositionLimit)

    def set_torque_limit(self, tl: int):
        """ Set the torque limit of the driver in terms of milliamps (mA).
        Torque is disabled after a timeout if the current drawn is over the
        given torque limit. Default torque limit is 65535.

        Args:
            id (int): The device ID of the driver.
            tl (int): New torque limit (mA)
        """
        self.set_variables([Index_Red.TorqueLimit, tl])
        self._post_sleep()

    def get_torque_limit(self):
        """ Get the torque limit from the driver in terms of milliamps (mA).

        Args:
            id (int): The device ID of the driver.

        Returns:
            list | None: Returns the list containing the torque limit, otherwise None.
        """
        return self.get_variables(Index_Red.TorqueLimit)[0]

    def set_velocity_limit(self, vl: int):
        """ Set the velocity limit for the motor output shaft in terms of RPM. The velocity limit
        applies only in velocity mode. Default velocity limit is 65535.

        Args:
            id (int): The device ID of the driver.
            vl (int): New velocity limit (RPM)
        """
        self.set_variables([Index_Red.VelocityLimit, vl])
        self._post_sleep()

    def get_velocity_limit(self):
        """ Get the velocity limit from the driver in terms of RPM.

        Args:
            id (int): The device ID of the driver.

        Returns:
            list | None: Returns the list containing the velocity limit, otherwise None.
        """
        return self.get_variables(Index_Red.VelocityLimit)[0]

    def set_position(self, sp: int):
        """ Set the desired setpoint for the position control in terms of encoder ticks.

        Args:
            id (int): The device ID of the driver.
            sp (int | float): Position control setpoint.
        """
        self.set_variables([Index_Red.PositionControlMode, 0],[Index_Red.SetPosition, sp])
        self._post_sleep()

    def get_position(self):
        """ Get the current position of the motor from the driver in terms of encoder ticks.

        Args:
            id (int): The device ID of the driver.

        Returns:
            list | None: Returns the list containing the current position, otherwise None.
        """
        return self.get_variables(Index_Red.PresentPosition)[0]
    
    def goTo(self, target_position, time_ = 0, maxSpeed = 0, accel = 0, 
             blocking: bool = False, encoder_tick_close_counter = 10):
        """
            # goTo: 

            Sets the target position in S Curve mode. Since this function controls motor in position, 
            device should be in Position Mode to use this func.

            If you just want to drive the motor to target point smoothly, you can use just the target_position parameter only.
            If other parameters are not given or they are 0 (time, maxSpeed, accel), it will use default speed and acceleration values based on the motor's RPM.
            
            If the setpoint, time, maxSpeed, and accel parameters are specified in a way that makes it impossible to reach the target in the given time, 
            the time variable will be ignored.

            If only the time is not provided, the movement will be executed according to the other given parameters.

            It is not necessary for the speed to reach the maxSpeed value during the movement. 
            The maxSpeed parameter is only a limitation. Due to the other given parameters, it might be impossible to reach the maxSpeed value during the movement. 
            Note: The motor's RPM value is defined as maxSpeed within the SMD.

            If blocking is True, the function will wait until the motor reaches the target position.
            If blocking is False, the function will return immediately.

            Args:
                id (int): The device ID of the driver.
                target_position (int | float): Position control setpoint.
                time (int | float): Time in seconds.
                maxSpeed (int | float): Maximum speed in RPM.
                accel (int | float): Acceleration in RPM/s.
                blocking (bool): If True, the function will wait until the motor reaches the target position.
                encoder_tick_close_counter (int): The number of encoder ticks that the motor should close enough to the target position to be considered reached.
        """

        self.set_variables([Index_Red.PositionControlMode, 1])
        self.set_variables([Index_Red.SCurveTime, time_],[Index_Red.SCurveMaxVelocity, maxSpeed],[Index_Red.ScurveAccel, accel])
        self.set_variables([Index_Red.SCurveSetpoint, target_position])

        self._post_sleep()

        while(blocking):
            if (abs(target_position - self.get_position(id)) <= encoder_tick_close_counter):
                break
        
    def goTo_ConstantSpeed(self, target_position, speed,
                           blocking: bool = False, encoder_tick_close_counter = 10):
        """
            # goTo_ConstantSpeed: 

            Sets the target position and sets the accel to max accel in S Curve mode. So velocity reaches given speed immediately.

            If blocking is True, the function will wait until the motor reaches the target position.
            If blocking is False, the function will return immediately.

            Args:
                id (int): The device ID of the driver.
                target_position (int | float): Position control setpoint.
                speed (int | float): Maximum speed in RPM.
                blocking (bool): If True, the function will wait until the motor reaches the target position.
                encoder_tick_close_counter (int): The number of encoder ticks that the motor should close enough to the target position to be considered reached.
        """
        
        self.set_variables([Index_Red.VelocityControlMode, 1])
        self.set_variables([Index_Red.SCurveMaxVelocity, speed],[Index_Red.ScurveAccel, MotorConstants.MAX_ACCEL])
        self.set_variables([Index_Red.SCurveSetpoint, target_position])

        self._post_sleep()

        while(blocking):
            if (abs(target_position - self.get_position()) <= encoder_tick_close_counter):
                break

    def set_velocity(self, sp: float, accel = 0):
        """ Set the desired setpoint for the velocity control in terms of RPM.

        Args:
            id (int): The device ID of the driver.
            sp (int | float): Velocity control setpoint.
            accel(float): sets the acceleration value for the velocity control in terms of (RPM/seconds). if accel is not given, it will be ignored. 
            So previously set accel value will be used.
            In initial SMD-RED Velocity Control Mode, accel will be set to MAX_ACCEL.

        Hint: 
            It can be used to set the acceleration value by "MotorConstants.MAX_ACCEL" to reach the target velocity immediately.
        """
        if accel == MotorConstants.MAX_ACCEL:
            accel = 0
            self.set_variables([Index_Red.SetVelocityAcceleration, accel])
            self.set_variables([Index_Red.SetVelocity, sp])

        elif accel == 0:
            self.set_variables([Index_Red.SetVelocity, sp])
        else:
            self.set_variables([Index_Red.SetVelocityAcceleration, accel])
            self.set_variables([Index_Red.SetVelocity, sp])
        
        self._post_sleep()

    def get_velocity(self):
        """ Get the current velocity of the motor output shaft from the driver in terms of RPM.

        Args:
            id (int): The device ID of the driver.

        Returns:
            list | None: Returns the list containing the current velocity, otherwise None.
        """
        return self.get_variables(Index_Red.PresentVelocity)[0]

    def set_torque(self, sp: float):
        """ Set the desired setpoint for the torque control in terms of milliamps (mA).

        Args:
            id (int): The device ID of the driver.
            sp (int | float): Torque control setpoint.
        """
        self.set_variables([Index_Red.SetTorque, sp])
        self._post_sleep()

    def get_torque(self):
        """ Get the current drawn from the motor from the driver in terms of milliamps (mA).

        Args:
            id (int): The device ID of the driver.

        Returns:
            list | None: Returns the list containing the current, otherwise None.
        """
        return self.get_variables(Index_Red.MotorCurrent)[0]

    def set_duty_cycle(self, pct: float):
        """ Set the duty cycle to the motor for PWM control mode in terms of percentage.
        Negative values will change the motor direction.

        Args:
            id (int): The device ID of the driver.
            pct (int | float): Duty cycle percentage.
        """
        self.set_variables([Index_Red.SetDutyCycle, pct])
        self._post_sleep()

    def get_analog_port(self):
        """ Get the ADC values from the analog port of the device with
        10 bit resolution. The value is in range [0, 4095].

        Args:
            id (int): The device ID of the driver.

        Returns:
            list | None: Returns the list containing the ADC conversion of the port, otherwise None.
        """
        return self.get_variables(Index_Red.AnalogPort)[0]

    def set_control_parameters_position(self, p=None, i=None, d=None, db=None, ff=None, ol=None):
        """ Set the control block parameters for position control mode.
        Only assigned parameters are written, None's are ignored. The default
        max output limit is 950.

        Args:
            id (int): The device ID of the driver.
            p (float): Proportional gain. Defaults to None.
            i (float): Integral gain. Defaults to None.
            d (float): Derivative gain. Defaults to None.
            db (float): Deadband (of the setpoint type). Defaults to None.
            ff (float): Feedforward. Defaults to None.
            ol (float): Maximum output limit. Defaults to None.
        """
        index_list = [Index_Red.PositionPGain, Index_Red.PositionIGain, Index_Red.PositionDGain, Index_Red.PositionDeadband, Index_Red.PositionFF, Index_Red.PositionOutputLimit]
        val_list = [p, i, d, db, ff, ol]
        self.set_variables(*[list(pair) for pair in zip(index_list, val_list) if pair[1] is not None])
        self._post_sleep()

    def get_control_parameters_position(self):
        """ Get the position control block parameters.

        Args:
            id (int): The device ID of the driver.

        Returns:
            list | None: Returns the list [P, I, D, FF, DB, OUTPUT_LIMIT], otherwise None.
        """

        return self.get_variables(*[Index_Red.PositionPGain, Index_Red.PositionIGain, Index_Red.PositionDGain, Index_Red.PositionDeadband, Index_Red.PositionFF, Index_Red.PositionOutputLimit])

    def set_control_parameters_velocity(self, p=None, i=None, d=None, db=None, ff=None, ol=None):
        """ Set the control block parameters for velocity control mode.
        Only assigned parameters are written, None's are ignored. The default
        max output limit is 950.

        Args:
            id (int): The device ID of the driver.
            p (float): Proportional gain. Defaults to None.
            i (float): Integral gain. Defaults to None.
            d (float): Derivative gain. Defaults to None.
            db (float): Deadband (of the setpoint type). Defaults to None.
            ff (float): Feedforward. Defaults to None.
            ol (float): Maximum output limit. Defaults to None.
        """
        index_list = [Index_Red.VelocityPGain, Index_Red.VelocityIGain, Index_Red.VelocityDGain, Index_Red.VelocityDeadband, Index_Red.VelocityFF, Index_Red.VelocityOutputLimit]
        val_list = [p, i, d, db, ff, ol]

        self.set_variables(*[list(pair) for pair in zip(index_list, val_list) if pair[1] is not None])
        self._post_sleep()

    def get_control_parameters_velocity(self):
        """ Get the velocity control block parameters.

        Args:
            id (int): The device ID of the driver.

        Returns:
            list | None: Returns the list [P, I, D, FF, DB, OUTPUT_LIMIT], otherwise None.
        """
        return self.get_variables(*[Index_Red.VelocityPGain, Index_Red.VelocityIGain, Index_Red.VelocityDGain, Index_Red.VelocityDeadband, Index_Red.VelocityFF, Index_Red.VelocityOutputLimit])

    def set_control_parameters_torque(self, p=None, i=None, d=None, db=None, ff=None, ol=None):
        """ Set the control block parameters for torque control mode.
        Only assigned parameters are written, None's are ignored. The default
        max output limit is 950.

        Args:
            id (int): The device ID of the driver.
            p (float): Proportional gain. Defaults to None.
            i (float): Integral gain. Defaults to None.
            d (float): Derivative gain. Defaults to None.
            db (float): Deadband (of the setpoint type). Defaults to None.
            ff (float): Feedforward. Defaults to None.
            ol (float): Maximum output limit. Defaults to None.
        """
        index_list = [Index_Red.TorquePGain, Index_Red.TorqueIGain, Index_Red.TorqueDGain, Index_Red.TorqueDeadband, Index_Red.TorqueFF, Index_Red.TorqueOutputLimit]
        val_list = [p, i, d, db, ff, ol]

        self.set_variables(*[list(pair) for pair in zip(index_list, val_list) if pair[1] is not None])
        self._post_sleep()

    def get_control_parameters_torque(self):
        """ Get the torque control block parameters.

        Args:
            id (int): The device ID of the driver.

        Returns:
            list | None: Returns the list [P, I, D, FF, DB, OUTPUT_LIMIT], otherwise None.
        """
        return self.get_variables(*[Index_Red.TorquePGain, Index_Red.TorqueIGain, Index_Red.TorqueDGain, Index_Red.TorqueDeadband, Index_Red.TorqueFF, Index_Red.TorqueOutputLimit])



def scan_red_devices(port:SerialPort):
    # burada herhangi bir nesne olusturmadan ping atmasi icin yeniden duzenlenecek.
    id_list = []
    for i in range(255):
        dev = Red(i, port)
        if dev.ping()== True:
            id_list.append(i)
    return id_list


def Red_set_variables_sync(parameter_index:Index_Red, port:SerialPort, id_val_pairs=[]):
        raise NotImplementedError()
        pass
        '''
        dev = Red(self.__class__._BROADCAST_ID)
        dev.vars[Index.Command].value(Commands.SYNC_WRITE)

        fmt_str = '<' + ''.join([var.type() for var in dev.vars[:6]])
        struct_out = list(struct.pack(fmt_str, *[var.value() for var in dev.vars[:6]]))

        fmt_str += 'B'
        struct_out += list(struct.pack('<B', int(index)))

        for pair in id_val_pairs:
            fmt_str += 'B'
            struct_out += list(struct.pack('<B', pair[0]))
            struct_out += list(struct.pack('<' + dev.vars[index].type(), pair[1]))

        struct_out[int(Index.PackageSize)] = len(struct_out) + dev.vars[Index.CRCValue].size()
        dev.vars[Index.CRCValue].value(CRC32.calc(struct_out))

        self.__write_bus(bytes(struct_out) + struct.pack('<' + dev.vars[Index.CRCValue].type(), dev.vars[Index.CRCValue].value()))
        self._post_sleep()
        '''





import serial

class SerialPort:
    def __init__(self, port_name, baudrate=921600, timeout=0.1, isTest:bool=False):
        self.port_name = port_name
        self.baudrate = baudrate
        self.timeout = timeout
        self.isTest = isTest
        if(self.isTest != True):
            self._ph = serial.Serial(port=self.port_name, baudrate=self.baudrate, timeout=self.timeout)

    def close_port(self):
        if self.isTest:
            print(f"virtual port destroyed.")
        else:
            if self._ph and self._ph.is_open:
                self._ph.reset_input_buffer()
                self._ph.reset_output_buffer()
                self._ph.close()
                print(f"Port '{self.port_name}' shut down.")
            else:
                print(f"Port '{self.port_name}' already closed.")

    def __del__(self):
        try:
            self.close_port()
        except Exception as e:
            print(f"Port yok edilirken bir hata oluştu: {e}")
        
    def _write_bus(self,data):
        if(self.isTest == True):
            print(list(data))
        else:
            self._ph.flushInput()
            self._ph.write(data)
        pass

    def _read_bus(self, size):
        if(self.isTest == False):
            return self._ph.read(size=size)
        else:
            print("Read Bus(TEST!)")
            return "True for test"

    def _no_timeout(self):
        notimeout = 50 #in seconds
        if(self.isTest):
            print(f"port timeout update is setted to {notimeout} (TEST!)")
        else:
            self._ph.timeout = notimeout

    def set_timeout(self, timeout):
        if(self.isTest):
            print(f"port timeout update to {timeout} (TEST!)")
        else:
            self._ph.timeout = timeout






import struct
from crccheck.crc import Crc32Mpeg2 as CRC32
import serial
import time
from packaging.version import parse as parse_version
import requests
import hashlib
import tempfile
from stm32loader.main import main as stm32loader_main
import enum
from smd.SMD_device import *


# enter here for extra commands: 
class Device_ExtraCommands(enum.IntEnum):
	# .......... 10
	EXTERNAL_TRAJECTORY_SETPOINT 	= 0x11,
	TUNE							= 0x12,
	# .......... 39


Index_Blue = enum.IntEnum('Index', [
	'Header',
	'DeviceID',
	'DeviceFamily',
	'PackageSize',
	'Command',
	'Status',
	'HardwareVersion',
	'SoftwareVersion',
	'Baudrate', #'WritableStart' = iBaudrate
	# user parameter start
	'OperationMode',
	'Enable',
	'CurrentSetting_Drive',
	'CurrentSetting_Hold',
	'Microstepping',
	'AutoStepInterpolation_enable',
	'AutoStepInterpolation_setting',
	'MaxAcceleration',
	'MaxDeceleration',
	'MaxSpeed',
	'MaxPosition',
	'MinPosition',
	'ExternalSetpoint_BufferSize',
	'ExternalSetpoint_PhaseDelay',
	'ExternalSetpoint_IntervalTime',
	'CurrentPosition',
	'CurrentSpeed',
	'CurrentAccel',
	'LimitSwitch_1',
	'LimitSwitch_2',
	'TargetVelocity',
	'TargetPosition',
	'DesiredTime',
	'DesiredAccel',
	'DesiredMaxSpeed',
	'Setpoint',
	# user parameter end
	'Config_TimeStamp',
	'Config_Description',
	'CRCValue',
], start=0)



def scan_blue_devices(port:SerialPort):
	device = Blue(0, port)
	available_devices = []

	for id in range(0,255):
		device._id = id
		if(device.ping()):
			available_devices.append(id)

	return available_devices


class Blue(SMD_Device):
	_PRODUCT_TYPE = 0xCA
	_PACKAGE_ESSENTIAL_SIZE = 6
	_STATUS_KEY_LIST = ['EEPROM', 'Software Version', 'Hardware Version']
	__RELEASE_URL = "https://api.github.com/repos/AAcrome-Smart-Motion-Devices/SMD-Blue-Firmware/releases/{version}"

	class Operation_Mode():
		Position_Internal_Trajectory = 0
		Position_External_Trajectory = 1
		Velocity = 2
	
	class microStepping():
		FULL_STEP 				= 0b0000
		FULL_STEP_71 			= 0b0001
		HALF_STEP_NON_CIRCULAR 	= 0b0010
		HALF_STEP 				= 0b0011
		_4_STEP					= 0b0100
		_8_STEP 				= 0b0101
		_16_STEP 				= 0b0110
		_32_STEP 				= 0b0111
		_64_STEP 				= 0b1000
		_128_STEP 				= 0b1001
		_256_STEP 				= 0b1010

	class autoStepInterpolation():
		_256_Interpolation 	= 0b00
		_128_Interpolation 	= 0b01
		_64_Interpolation	= 0b10
		_32_Interpolation	= 0b11


	def __init__(self, ID, port:SerialPort) -> bool:
		self.__ack_size = 0
		if ID > 254 or ID < 0:
			raise ValueError("Device ID can not be higher than 254 or lower than 0!")
		device_special_data = [
            Data_(Index_Blue.Header, 'B', False, 0x55),
            Data_(Index_Blue.DeviceID, 'B'),
			Data_(Index_Blue.DeviceFamily, 'B'),
            Data_(Index_Blue.PackageSize, 'B'),
            Data_(Index_Blue.Command, 'B'),
			Data_(Index_Blue.Status, 'B'),
            Data_(Index_Blue.HardwareVersion, 'I'),
            Data_(Index_Blue.SoftwareVersion, 'I'),
            Data_(Index_Blue.Baudrate, 'I'),
			# user parameter starts
			Data_(Index_Blue.OperationMode, 'B'),
			Data_(Index_Blue.Enable, 'B'),
			Data_(Index_Blue.CurrentSetting_Drive, 'B'),
			Data_(Index_Blue.CurrentSetting_Hold, 'B'),
			Data_(Index_Blue.Microstepping, 'B'),
			Data_(Index_Blue.AutoStepInterpolation_enable, 'B'),
			Data_(Index_Blue.AutoStepInterpolation_setting, 'B'),
			Data_(Index_Blue.MaxAcceleration, 'f'),
			Data_(Index_Blue.MaxDeceleration, 'f'),
			Data_(Index_Blue.MaxSpeed, 'f'),
			Data_(Index_Blue.MaxPosition, 'i'),
			Data_(Index_Blue.MinPosition, 'i'),
			Data_(Index_Blue.ExternalSetpoint_BufferSize, 'H'),
			Data_(Index_Blue.ExternalSetpoint_PhaseDelay, 'H'),
			Data_(Index_Blue.ExternalSetpoint_IntervalTime, 'I'),
			Data_(Index_Blue.CurrentPosition, 'd'),
			Data_(Index_Blue.CurrentSpeed, 'd'),
			Data_(Index_Blue.CurrentAccel, 'd'),
			Data_(Index_Blue.LimitSwitch_1, 'B'),
			Data_(Index_Blue.LimitSwitch_2, 'B'),
			Data_(Index_Blue.TargetVelocity, 'f'),
			Data_(Index_Blue.TargetPosition, 'i'),
			Data_(Index_Blue.DesiredTime, 'f'),
			Data_(Index_Blue.DesiredAccel, 'f'),
			Data_(Index_Blue.DesiredMaxSpeed, 'f'),
			Data_(Index_Blue.Setpoint, 'f'),
			# user parameter end
			Data_(Index_Blue.Config_TimeStamp, 'Q'),
			Data_(Index_Blue.Config_Description, '100s'),
            Data_(Index_Blue.CRCValue, 'I'),
        ]
		super().__init__(ID, self._PRODUCT_TYPE, device_special_data, port)
		self._vars[Index_Blue.DeviceID].value(ID)

	# user start for extra commands.
	#def command(self): 

	def __del__(self):
		pass

	def get_latest_fw_version(self):
		""" Get the latest firmware version from the Github servers.

		Returns:
			String: Latest firmware version
		"""
		response = requests.get(url=self.__class__.__RELEASE_URL.format(version='latest'))
		if (response.status_code in [200, 302]):
			return (response.json()['tag_name'])

	def update_fw_version(self, version=''):
		""" Update firmware version with respect to given version string.

		Args:
			id (int): The device ID of the driver
			version (str, optional): Desired firmware version. Defaults to ''.

		Returns:
			Bool: True if the firmware is updated
		"""

		fw_file = tempfile.NamedTemporaryFile("wb+",delete=False)
		if version == '':
			version = 'latest'
		else:
			version = 'tags/' + version

		response = requests.get(url=self.__class__.__RELEASE_URL.format(version=version))
		if response.status_code in [200, 302]:
			assets = response.json()['assets']

			fw_dl_url = None
			md5_dl_url = None
			for asset in assets:
				if '.bin' in asset['name']:
					fw_dl_url = asset['browser_download_url']
				elif '.md5' in asset['name']:
					md5_dl_url = asset['browser_download_url']

			if None in [fw_dl_url, md5_dl_url]:
				raise Exception("Could not found requested firmware file! Check your connection to GitHub.")

			#  Get binary firmware file
			md5_fw = None
			response = requests.get(fw_dl_url, stream=True)
			if (response.status_code in [200, 302]):
				fw_file.write(response.content)
				md5_fw = hashlib.md5(response.content).hexdigest()
			else:
				raise Exception("Could not fetch requested binary file! Check your connection to GitHub.")

			#  Get MD5 file
			response = requests.get(md5_dl_url, stream=True)
			if (response.status_code in [200, 302]):
				md5_retreived = response.text.split(' ')[0]
				if (md5_fw == md5_retreived):

					# Put the driver in to bootloader mode
					self.enter_bootloader()
					time.sleep(0.1)

					# Close serial port
					serial_settings = self._port._ph.get_settings()
					self._port._ph.close()

					# Upload binary
					args = ['-p', self._port._ph.portstr, '-b', str(115200), '-e', '-w', '-v', fw_file.name]
					stm32loader_main(*args)

					# Delete uploaded binary
					if (not fw_file.closed):
						fw_file.close()

					# Re open port to the user with saved settings
					self._port._ph.apply_settings(serial_settings)
					self._port._ph.open()
					return True

				else:
					raise Exception("MD5 Mismatch!")
			else:
				raise Exception("Could not fetch requested MD5 file! Check your connection to GitHub.")
		else:
			raise Exception("Could not found requested firmware files list! Check your connection to GitHub.")
		
	def enable_torque(self, en: bool):
		""" Enable power to the motor of the driver.

    	Args:
    	    id (int): The device ID of the driver
    	    en (bool): Enable. True enables the torque.
    	"""

		self.set_variables([Index_Blue.Enable, en])
		self._post_sleep()

	def set_microstepping(self, microstepping:int, auto_stepping_enable:bool = True, auto_stepping_interpolation:int= autoStepInterpolation._256_Interpolation):
		if (microstepping > self.microStepping._256_STEP or microstepping < 0):
			raise "microstepping is not valid."
		if (auto_stepping_interpolation > self.autoStepInterpolation._32_Interpolation or auto_stepping_interpolation < self.autoStepInterpolation._256_Interpolation):
			raise "auto_stepping_interpolation is not valid."
		
		self.set_variables([Index_Blue.Microstepping, microstepping], [Index_Blue.AutoStepInterpolation_enable, auto_stepping_enable], [Index_Blue.AutoStepInterpolation_setting, auto_stepping_interpolation])
		self._post_sleep()


	def set_config_timeStamp(self):
		epoch_seconds = int(time.time())
		self.set_variables([Index_Blue.Config_TimeStamp, epoch_seconds])
		self._post_sleep()
		
	def set_config_description(self, description:str):
		if len(description) >= 100:
			text = description[:99] + '\0'
		else:
			text = description + '\0'
			text = text.ljust(100, ' ')
		text = text.encode('ascii')  # veya utf-8 eğer uyumluysa

		self.set_variables([Index_Blue.Config_Description, text])
		self._post_sleep()


## this is master for connected devices.
import struct
from crccheck.crc import Crc32Mpeg2 as CRC32
import time
import enum
from smd.serial_port import SerialPort
'''
COMMUNICATION PACKAGE => 
HEADER, ID, DEVICE_FAMILY, PACKAGE_SIZE, COMMAND, STATUS, .............. DATA ................. , CRC
'''

SMD_SERIAL_HEADER = 0x55
SMD_PING_PACKAGE_SIZE = 10

#Classical Device Indexes
Index_Device_Classical = enum.IntEnum('Index', [
	'Header',
	'DeviceID',
    'DeviceFamily',
	'PackageSize',
	'Command',
    'Status',
	'HardwareVersion',
	'SoftwareVersion',
    'Baudrate',
], start=0)

#Classical_Commands
class Device_Commands(enum.IntEnum):
    ACK = 0x80
    SYNC = 0x40

    PING = 0x00
    WRITE = 0x01
    READ = 0x02
    EEPROM_SAVE = 0x03
    ERROR_CLEAR = 0x04
    REBOOT = 0x05
    EEPROM_RESET = 0x06
    BL_JUMP = 0x07
    ENTER_CONFIGURATION = 0x08
    ENTER_OPERATION = 0x09
    ERROR = 0x10

    ENTER_CONFIGURATION_ACK = 0x80 | 0x08       # ACK | ENTER_CONFIGURATION
    ENTER_OPERATION_ACK = 0x80 | 0x09           # ACK | ENTER_OPERATION
    WRITE_ACK = 0x80 | 0x01                     # ACK | WRITE
    WRITE_SYNC = 0x40 | 0x01                    # SYNC | WRITE
    EEPROM_SAVE_ACK = 0x80 | 0x03               # ACK | EEPROM_WRITE
    EEPROM_SAVE_SYNC = 0x40 | 0x03              # SYNC | EEPROM_WRITE

def set_variables_directly(header:int, device_family:int, id:int, status:int=0, *idx_val_pairs, ack = False, port:SerialPort):
        # returns : did ACK come?
        port._ph.flushInput()

        fmt_str = '<BBBBBB'
        var_count = 0
        size = 0
        for one_pair in idx_val_pairs:
            try:
                if len(one_pair) != 2:
                    raise ValueError(f"{one_pair} more than a pair! It is not a pair")
                else:
                    fmt_str += ('B' + self._vars[one_pair[0]].type())
                    var_count+=1
                    size += (1 + self._vars[one_pair[0]].size())
            except:
                raise ValueError(f"{one_pair} is not proper pair")
        
        flattened_list = [item for sublist in idx_val_pairs for item in sublist]

        struct_out = list(struct.pack(fmt_str, *[header, id, device_family, size + SMD_PING_PACKAGE_SIZE, Device_Commands.WRITE, status, *flattened_list]))
        struct_out = bytes(struct_out) + struct.pack('<' + 'I', CRC32.calc(struct_out))
        ack_size = SMD_PING_PACKAGE_SIZE


        port._ph.write(struct_out)
        raise NotImplementedError()

        if _read_ack():
            return True
        else:
            return False


class SMD_Device():
    SMD_SERIAL_HEADER = 0x55
    _BROADCAST_ID = 0xFF

    _BATCH_ID = 254
    def __init__(self, id, device_family, variables, port:SerialPort):
        self._port = port
        self._header = self.SMD_SERIAL_HEADER
        self._id = id
        self._device_family = device_family
        self._vars = variables
        self._ack_size = 0
        self.__post_sleep = 0.01
        self.__device_init_sleep = 3
        self.write_ack_enable = False

    def enable_get_ack(self):
        self.write_ack_enable = True

    def _init_sleep(self):
        time.sleep(self.__device_init_sleep)

    def _post_sleep(self):
        time.sleep(self.__post_sleep)
    
    def _write_port(self, data):
        self._port._write_bus(data)

    def _read_port(self, size) -> bytes:
        return self._port._read_bus(size=size)
    
    def _parse_received(self, data):
        id = data[Index_Device_Classical.DeviceID]
        data = data[4:-4]
        fmt_str = '<'

        i = 0
        while i < len(data):
            fmt_str += 'B' + self._vars[data[i]].type()
            i += self._vars[data[i]].size() + 1

        unpacked = list(struct.unpack(fmt_str, data))
        grouped = zip(*(iter(unpacked),) * 2)
        for group in grouped:
            self._vars[group[0]].value(group[1])
    
    def _read_ack(self) -> bool:
        ret = self._read_port(self._ack_size)
        if (ret == "True for test"):
            return True
        if(ret==None):
            return False
        if len(ret) == self._ack_size:
            if (CRC32.calc(ret[:-4]) == struct.unpack('<I', ret[-4:])[0]):
                if ret[Index_Device_Classical.PackageSize] > SMD_PING_PACKAGE_SIZE:
                    self._parse_received(ret)
                    return True
                else:
                    return True # ping islemi ve WRITE_ACK icin.
            else:
                return False
        else:
            return False
        
    def _read_var_no_timeout(self):
        self._port._no_timeout()
        ack_flag = self._read_ack()
        self._port.set_timeout(0.01)
        if ack_flag:
            return True
        else:
            return False 
        
    def _pure_command_send(self, command:int):
        fmt_str = '<BBBBBB'
        struct_out = list(struct.pack(fmt_str, *[self._header, self._id, self._device_family, SMD_PING_PACKAGE_SIZE, command, 0]))
        struct_out = bytes(struct_out) + struct.pack('<I', CRC32.calc(struct_out))
        self._write_port(struct_out)
        return struct_out

    def ping(self):
        self._pure_command_send(Device_Commands.PING)
        self._ack_size = SMD_PING_PACKAGE_SIZE
        if self._read_ack():
            return True
        else:
            return False
    
    def get_variables(self, *indexes):
        self._ack_size = 0
        fmt_str = '<BBBBBB'+'B'*len(indexes)

        struct_out = list(struct.pack(fmt_str, *[self._header, self._id, self._device_family ,len(indexes) + SMD_PING_PACKAGE_SIZE, Device_Commands.READ, 0, *indexes]))
        struct_out = bytes(struct_out) + struct.pack('<' + 'I', CRC32.calc(struct_out))
        for i in indexes:
            self._ack_size += (self._vars[int(i)].size() + 1)
        self._ack_size += SMD_PING_PACKAGE_SIZE

        self._write_port(struct_out)

        if self._read_ack():
            return [self._vars[index].value() for index in indexes]
        else:
            return [None]

    def set_variables(self, *idx_val_pairs, ack = False):
        # returns : did ACK come?
        fmt_str = '<BBBBBB'
        var_count = 0
        size = 0
        for one_pair in idx_val_pairs:
            try:
                if len(one_pair) != 2:
                    raise ValueError(f"{one_pair} more than a pair! It is not a pair")
                else:
                    fmt_str += ('B' + self._vars[one_pair[0]].type())
                    var_count+=1
                    size += (1 + self._vars[one_pair[0]].size())
            except:
                raise ValueError(f"{one_pair} is not proper pair")
        
        flattened_list = [item for sublist in idx_val_pairs for item in sublist]

        struct_out = list(struct.pack(fmt_str, *[self._header, self._id, self._device_family, size + SMD_PING_PACKAGE_SIZE, Device_Commands.WRITE, 0, *flattened_list]))
        struct_out = bytes(struct_out) + struct.pack('<' + 'I', CRC32.calc(struct_out))
        self._ack_size = SMD_PING_PACKAGE_SIZE

        self._write_port(struct_out)
        if(self.write_ack_enable):
            if self._read_ack():
                return True
            else:
                return False
        return False
        
    def reboot(self):
        self._pure_command_send(Device_Commands.REBOOT)
	
    def eeprom_save(self):
        self._pure_command_send(Device_Commands.EEPROM_SAVE)
        
    def factory_reset(self, ack=False):
        self._pure_command_send(Device_Commands.EEPROM_RESET)

    def enter_bootloader(self):
        self._pure_command_send(Device_Commands.BL_JUMP)

    def enter_operation(self):
        self._pure_command_send(Device_Commands.ENTER_OPERATION)

    def enter_configuration(self):
        self._pure_command_send(Device_Commands.ENTER_CONFIGURATION)

    def get_driver_info(self):
        """ Get hardware and software versions from the driver

        Args:
            id (int): The device ID of the driver.

        Returns:
            dict | None: Dictionary containing versions or None.
        """
        st = dict()
        data = self.get_variables([Index_Device_Classical.HardwareVersion, Index_Device_Classical.SoftwareVersion])
        if data is not None:
            ver = list(struct.pack('<I', data[0]))
            st['HardwareVersion'] = "v{1}.{2}.{3}".format(*ver[::-1])
            ver = list(struct.pack('<I', data[1]))
            st['SoftwareVersion'] = "v{1}.{2}.{3}".format(*ver[::-1])

            self.__driver_list[id]._config = st
            return st
        else:
            return None
        
    def update_driver_id(self, id: int, id_new: int):
        """ Update the device ID of the driver

        Args:
            id (int): The device ID of the driver
            id_new (int): New device ID

        Raises:
            ValueError: Current or updating device IDs are not valid
        """
        if (id < 0) or (id > 254):
            raise ValueError("{} is not a valid ID!".format(id))

        if (id_new < 0) or (id_new > 254):
            raise ValueError("{} is not a valid ID argument!".format(id_new))
        
        self.set_variables([Index_Device_Classical.DeviceID, id_new])
        self._post_sleep()
        
        self.eeprom_save(id_new)
        self._post_sleep()
        self.reboot(id)
        
    def get_all_variable(self):
        for i in range(0, len(self._vars), 10):
            j = i
            k = min(i + 9, len(self._vars) - 1)  # Son grupta sınırlamayı sağlar
            index_list = list(range(j, k + 1))
            self.read_var(*index_list) # her birisi maksimum data sayisiymis gibi dusunerek yazarsak 4 byte olur. her bir pakette 10 adet alsin. maksimuma vurmak istemedigimizden dolayi.

class Data_():
    def __init__(self, index, var_type, rw=True, value = 0):
        self.__index = index
        self.__type = var_type
        self.__size  = struct.calcsize(self.__type)
        self.__value = value
        self.__rw = rw

    def value(self, value=None):
        if value is None:
            return self.__value
        elif self.__rw:
            self.__value = struct.unpack('<' + self.__type, struct.pack('<' + self.__type, value))[0]

    def index(self) ->enum.IntEnum:
        return self.__index
    
    def writeable(self) -> bool:
        return self.__rw

    def size(self) -> int:
        return self.__size
	
    def type(self) -> str:
        return self.__type



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="acrome-smd-beta",
    version="0.0.4",
    author="BeratComputer",
    author_email="beratdogan@acrome.net",
    description="Python library for interfacing with Acrome Robotic Arm Joint BLDC Motor Controllers (Beta Version) \n \n This Python library provides an easy-to-use interface for communication and control of BLDC motor controllers used in Acrome robotic arm joints. It is designed to simplify the integration of Acrome’s robotic joint actuators into custom applications, allowing developers and researchers to focus on building advanced robotic systems without dealing with low-level communication details.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Acrome-Smart-Motion-Devices/python-library-new",
    project_urls={
        "Bug Tracker": "https://github.com/Acrome-Smart-Motion-Devices/python-library/issues",
        },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(exclude=['tests', 'test']),
    install_requires=["pyserial>=3.5", "stm32loader>=0.5.1", "crccheck>=1.3.0", "requests>=2.31.0", "packaging>=23.2"],
    python_requires=">=3.7"
)